#! /usr/bin/env python3

"""proteosushi_gui.py: displays the GUI and connects it to the main program"""

import importlib.resources as pkg_resources
import os
import sys
from time import time
from PyQt5 import QtCore
from PyQt5.QtWidgets import (QApplication, QWidget, QMainWindow, QPushButton, 
                             QMessageBox, QLineEdit, QLabel, QGroupBox, 
                             QGridLayout, QVBoxLayout, QFileDialog, QCheckBox,
                             QRadioButton, QButtonGroup, QComboBox)
from PyQt5.QtCore import pyqtSlot, QSize, Qt, QRunnable, QObject, QThreadPool, pyqtSignal
from PyQt5.QtGui import QIcon
try:
    from .combine_intensities import parse_output, rollup
    from . import lib
    from .parse_proteome import parse_proteome
    from .proteoSushi_constants import cleave_rules
except ImportError:
    from combine_intensities import parse_output, rollup
    import lib
    from parse_proteome import parse_proteome
    from proteoSushi_constants import cleave_rules


class WorkerSignals(QObject):
    """Class that contains the signal sent at the end of the worker thread"""
    end = pyqtSignal()


class Worker(QRunnable):
    """Class for the backend worker thread"""

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()    


    @pyqtSlot()
    def run(self):
        """Initializes the runner function"""
        # Retrieve args/kwargs here; and fire processing using them
        self.fn(*self.args, **self.kwargs)
        self.signals.end.emit()

class App(QMainWindow):

    def __init__(self, multithread: bool):
        super().__init__()
        self.title = "ProteoSushi"
        self.left = 10
        self.top = 10
        self.width = 640
        self.height = 480
        self.multithread = multithread
        self.centralWidget = QWidget()
        self.setCentralWidget(self.centralWidget)
        self.setWindowTitle("ProteoSushi")
        self.setWindowIcon(QIcon("ProteoSushi_icon.png"))
        self.species_id_dict = self.__make_species_dict()
        self.species_name_dict = {v:k for k, v in self.species_id_dict.items()}
        self.threadpool = QThreadPool()
        self.initUI()
        
    def initUI(self):
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)


        self.file_chooser_label = QLabel("Types of ProteoSushi Files", self)

        self.maxquant_RB = QRadioButton("MaxQuant", self)
        self.maxquant_RB.toggled.connect(self.check_MQ_RB)
        self.maxquant_RB.setToolTip("Use the txt output folder from the MaxQuant Search Engine")
        self.maxquant_button = QPushButton("MaxQuant txt Folder")
        self.maxquant_button.setHidden(True)
        self.maxquant_button.clicked.connect(self.onClickMQbutton)
        self.maxquant_filepath = QLabel("[Filepath]", self)
        self.maxquant_filepath.setHidden(True)

        self.mascot_RB = QRadioButton("Mascot", self)
        self.mascot_RB.toggled.connect(self.check_mascot_RB)
        self.mascot_RB.setToolTip("Use the output from the Mascot Search Engine")
        self.mascot_button = QPushButton("Mascot File")
        self.mascot_button.setHidden(True)
        self.mascot_button.clicked.connect(self.onClickMascotButton)
        self.mascot_filepath = QLabel("[Filepath]", self)
        self.mascot_filepath.setHidden(True)

        self.generic_RB = QRadioButton("Generic", self)
        self.generic_RB.toggled.connect(self.check_generic_RB)
        self.generic_RB.setToolTip("Use the output from any other search engine")
        self.generic_button = QPushButton("Generic File")
        self.generic_button.setHidden(True)
        self.generic_button.clicked.connect(self.onClickGenericButton)
        self.generic_filepath = QLabel("[Filepath]", self)
        self.generic_filepath.setHidden(True)

        self.search_engine_group = QButtonGroup()
        self.search_engine_group.addButton(self.maxquant_RB)
        self.search_engine_group.addButton(self.mascot_RB)
        self.search_engine_group.addButton(self.generic_RB)

        self.PTM_CBs = list()

        self.proteome_filepath_button = QPushButton("Uniprot Proteome FASTA")
        self.proteome_filepath_button.clicked.connect(self.on_click_proteome_button)
        self.proteome_filepath = QLabel("[Filepath]", self)

        self.output_filepath_button = QPushButton("Output Name and Location")
        self.output_filepath_button.clicked.connect(self.on_click_output_button)
        self.output_filepath = QLabel("[Filepath]", self)

        self.options_label = QLabel("Options", self)

        self.localization_checkbox = QCheckBox("Localization Threshold", self)
        self.localization_checkbox.stateChanged.connect(self.check_localization_checkbox)
        self.localization_checkbox.setHidden(True)
        self.localization_checkbox.setToolTip("Set a threshold for the localization probability for all PTM sites")
        self.localization_edit = QLineEdit(self)
        self.localization_edit.setHidden(True)
        self.localization_edit.setToolTip("Set a threshold for the localization probability for all PTM sites")

        self.target_checkbox = QCheckBox("Use Target Genes", self)
        self.target_checkbox.stateChanged.connect(self.checkTargetBox)
        self.target_checkbox.setToolTip("Use a list of genes that will be prioritized given multiple matches")
        self.target_button = QPushButton("Target Gene List")
        self.target_button.setHidden(True)
        self.target_button.clicked.connect(self.onClickTargetButton)
        self.target_filepath = QLabel("[Filepath]", self)
        self.target_filepath.setHidden(True)

        self.quant_CB = QCheckBox("Use Quantitation Values", self)
        self.quant_CB.stateChanged.connect(self.check_quant_CB)
        self.quant_CB.setToolTip("Whether to use the Intensity/Quantitation values from the PSMs")
        self.sum_RB = QRadioButton("Sum Peaks", self)
        self.sum_RB.setHidden(True)
        self.average_RB = QRadioButton("Average Peaks", self)
        self.average_RB.setHidden(True)

        self.quant_method_group = QButtonGroup()
        self.quant_method_group.addButton(self.sum_RB)
        self.quant_method_group.addButton(self.average_RB)

        self.uniprot_annot_CB = QCheckBox("Annotate with Uniprot", self)
        self.uniprot_annot_CB.setToolTip("Whether to add in annotation from Uniprot, like subcellular location, secondary structure, etc.")

        self.species_id_label = QLabel("Species ID", self)
        self.species_id_label.setToolTip("The species ID for the species from the data provided (e.g. 9606)")
        self.species_id_edit = QLineEdit(self)
        self.species_id_edit.editingFinished.connect(self.update_species_name)
        self.species_id_edit.setToolTip("The species ID for the species from the data provided (e.g. 9606)")
        self.species_name_label = QLabel("", self)

        self.max_missed_label = QLabel("Max Missed Cleavages", self)
        self.max_missed_label.setToolTip("The maximum allowed missed cleavages for a given peptide")
        self.max_missed_edit = QLineEdit(self)
        self.max_missed_edit.setToolTip("The maximum allowed missed cleavages for a given peptide")

        self.fdr_CB = QCheckBox("FDR Threshold", self)
        self.fdr_CB.stateChanged.connect(self.check_fdr_CB)
        self.fdr_CB.setToolTip("[OPTIONAL] The threshold for pep_expect column for Mascot or PEP column for Maxquant.\nMust be between 0 and 1, but can be left blank.")
        self.fdr_edit = QLineEdit(self)  # TODO: Error check for a number
        self.fdr_edit.setHidden(True)
        self.fdr_edit.setToolTip("[OPTIONAL] The threshold for pep_expect column for Mascot or PEP column for Maxquant.\nMust be between 0 and 1, but can be left blank.")

        self.protease_label = QLabel("Protease used in sample digestion", self)
        self.protease_label.setToolTip("The protease used to digest the sample\nExamples include: trypsin/p, trypsin!p, lys-c, asp-n, asp-ne, lys-n")
        self.protease_combo_box = QComboBox()
        self.protease_combo_box.setToolTip("The protease used to digest the sample\nExamples include: trypsin/p, trypsin!p, lys-c, asp-n, asp-ne, lys-n")
        self.protease_combo_box.addItems(["trypsin/p", "trypsin!p", "lys-c", "asp-n", "asp-ne", "lys-n"])
        self.protease_combo_box.setEditable(True)


        self.run_button = QPushButton("Rollup!")
        self.run_button.setHidden(False)
        self.run_button.clicked.connect(self.onClickRunButton)

        self.createGridLayout()
        windowLayout = QVBoxLayout(self.centralWidget)
        windowLayout.addWidget(self.horizontalGroupBox)
        windowLayout.setAlignment(Qt.AlignTop)
        self.setLayout(windowLayout)
        
        self.show()
    
    def check_MQ_RB(self, state):
        if self.maxquant_RB.isChecked():
            self.maxquant_filepath.setHidden(False)
            self.maxquant_button.setHidden(False)
            self.localization_checkbox.setHidden(False)
            if self.localization_checkbox.isChecked():
                self.localization_edit.setHidden(False)
        else:
            self.maxquant_filepath.setHidden(True)
            self.maxquant_button.setHidden(True)
            self.localization_checkbox.setHidden(True)
            self.localization_edit.setHidden(True)

    def check_mascot_RB(self, state):
        if self.mascot_RB.isChecked():
            self.mascot_filepath.setHidden(False)
            self.mascot_button.setHidden(False)
        else:
            self.mascot_filepath.setHidden(True)
            self.mascot_button.setHidden(True)
    
    def check_generic_RB(self, state):
        if self.generic_RB.isChecked():
            self.generic_filepath.setHidden(False)
            self.generic_button.setHidden(False)
            self.localization_checkbox.setHidden(False)
            if self.localization_checkbox.isChecked():
                self.localization_edit.setHidden(False)
        else:
            self.generic_filepath.setHidden(True)
            self.generic_button.setHidden(True)
            self.localization_checkbox.setHidden(True)
            self.localization_edit.setHidden(True)

    def checkTargetBox(self, state):
        if state == QtCore.Qt.Checked:
            self.target_filepath.setHidden(False)
            self.target_button.setHidden(False)
        else:
            self.target_filepath.setHidden(True)
            self.target_button.setHidden(True)
    
    def check_quant_CB(self, state):
        if state == QtCore.Qt.Checked:
            self.average_RB.setHidden(False)
            self.sum_RB.setHidden(False)
        else:
            self.average_RB.setHidden(True)
            self.sum_RB.setHidden(True)
    
    def check_fdr_CB(self, state):
        if state == QtCore.Qt.Checked:
            self.fdr_edit.setHidden(False)
        else:
            self.fdr_edit.setHidden(True)

    def check_localization_checkbox(self, state):
        if state == QtCore.Qt.Checked and (self.maxquant_RB.isChecked() or self.generic_RB.isChecked()):
            self.localization_edit.setHidden(False)
        else:
            self.localization_edit.setHidden(True)

    def update_species_name(self):
        if not self.species_id_edit.text(): 
            self.species_name_label.setStyleSheet("background-color : white")
            self.species_name_label.setText("")
        elif self.species_id_edit.text() in self.species_id_dict:
            self.species_name_label.setStyleSheet("background-color : white")
            self.species_name_label.setText(self.species_id_dict[self.species_id_edit.text()][2:])
        elif f"N={self.species_id_edit.text()}" in self.species_name_dict:
            self.species_name_label.setStyleSheet("background-color : white")
            self.species_name_label.setText(self.species_id_edit.text())
            self.species_id_edit.setText(self.species_name_dict[f"N={self.species_id_edit.text()}"])
        else:
            self.species_name_label.setText("Not a valid species ID")
            self.species_name_label.setStyleSheet("background-color : red")

    def checkRunButton(self, MQcheckState, MQfilepath: str, mascotCheckState, mascot_filepath: str, genericCheckState, generic_filepath: str):
        """Checks to see whether the run button is available. must have a search engine output and missed cleavages, etc
        Arguments:
            MQcheckState {QtCore.Qt.Checked} -- whether the box is checked
            MQfilepath {str} -- the filepath for maxquant output
            mascotCheckState {QtCore.Qt.Checked} -- whether the box is checked
            mascot_filepath {str} -- the filepath for maxquant output
            genericCheckState {QtCore.Qt.Checked} -- whether the box is checked
            generic_filepath {str} -- the filepath for maxquant output
        """
        # TODO: update to require species, etc.
        if (((MQcheckState == QtCore.Qt.Checked and os.path.exists(MQfilepath)) or
            (mascotCheckState == QtCore.Qt.Checked and os.path.exists(mascot_filepath)) or
            (genericCheckState == QtCore.Qt.Checked and os.path.exists(generic_filepath))) and
            os.path.exists(self.proteome_filepath.text())):
            self.run_button.setHidden(False)
        else:
            self.run_button.setHidden(True)


    def __make_species_dict(self):
        species_id_dict = dict()
        with pkg_resources.open_text(lib, "spec_list_fixed.tsv") as spec_list:
        #with pkg_resources.open_text(__package__, "spec_list_fixed.tsv") as spec_list:
            for line in spec_list:
                split_line = line.strip().split('\t')
                species_id_dict[split_line[0]] = split_line[1]
        return species_id_dict



    def openCSVFileNameDialog(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        fileName, _ = QFileDialog.getOpenFileName(self,"Choose the Search Engine Output", os.getcwd(),"CSV Files (*.csv);;All Files (*)", options=options)
        return fileName

    def openTXTFileNameDialog(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        fileName, _ = QFileDialog.getOpenFileName(self,"Choose the Gene Name File", os.getcwd(),"TXT Files (*.txt);;All Files (*)", options=options)
        return fileName

    def open_directory_dialog(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        dir_name = str(QFileDialog.getExistingDirectory(self, "Select Directory", os.getcwd(), options=options))
        return dir_name
    
    def openFASTAFileNameDialog(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        fileName, _ = QFileDialog.getOpenFileName(self,"Choose the Proteome FASTA file", os.getcwd(),"FASTA Files (*.fasta);;All Files (*)", options=options)
        return fileName
    
    def write_output_dialog(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        fileName, _ = QFileDialog.getSaveFileName(self,"Choose the name and location of the output", os.getcwd(),"CSV Files (*.csv);;All Files (*)", options=options)
        return fileName


    def convert_PTM_list(self, PTM_CBs:list) -> list:
        """changes the list of checkboxes into chosen ptms for analysis

        Arguments:
            PTM_CBs {list} -- a list of QCheckBoxes for each possible PTM
        Returns:
            list -- a list of strings of PTM names for analysis
        """
        PTM_str_list = list()
        for ptm_cb in PTM_CBs:
            if ptm_cb.isChecked():
                PTM_str_list.append(ptm_cb.text())
        return PTM_str_list

    @pyqtSlot()
    def onClickMQbutton(self):
        self.statusBar().showMessage("Choose the MaxQuant output FOLDER")
        self.statusBar().setStyleSheet("background-color : white")
        filename = self.open_directory_dialog()
        #if self.maxquant_filepath.text() != "":
        if os.path.exists(filename):
            self.maxquant_filepath.setText(filename)
            missed_cleavages, protease, PTMs = parse_output("maxquant", filename)
            self.max_missed_edit.setText(str(missed_cleavages))
            self.protease_combo_box.setEditText(protease)
            #Remove the previous PTM checkboxes (if there were any)
            if self.PTM_CBs != []:
                for cb in self.PTM_CBs:
                    self.layout.removeWidget(cb)
                    cb.deleteLater()

            # Adds the new checkboxes to the list
            self.PTM_CBs = []
            for ptm in PTMs:
                self.PTM_CBs.append(QCheckBox(ptm, self))
            # Removes the label
            i = 0
            if not self.layout.itemAtPosition(5, i) is None:
                label_to_remove = self.layout.itemAtPosition(5, i).widget()
                self.layout.removeWidget(label_to_remove)
                label_to_remove.deleteLater()

            # Inserts the new PTM checkboxes
            self.layout.addWidget(QLabel("PTMs for Current Analysis"), 5, i)
            for widget in self.PTM_CBs:
                self.layout.addWidget(widget, 6, i)
                i += 1
        self.statusBar().showMessage("")
            
    
    @pyqtSlot()
    def onClickMascotButton(self):
        self.statusBar().showMessage("Choose the Mascot output file")
        self.statusBar().setStyleSheet("background-color : white")
        filename = self.openCSVFileNameDialog()
        #if self.mascot_filepath.text() != "":
        if os.path.exists(filename):
            self.mascot_filepath.setText(filename)
            missed_cleavages, protease, PTMs = parse_output("mascot", filename)
            if missed_cleavages == -5:
                self.statusBar().showMessage("Invalid Mascot file")
                self.statusBar().setStyleSheet("background-color : red")
                self.mascot_filepath.setText("")
                return
            self.max_missed_edit.setText(str(missed_cleavages))
            self.protease_combo_box.setEditText(protease)
            #Remove the previous PTM checkboxes (if there were any)
            if self.PTM_CBs != []:
                for cb in self.PTM_CBs:
                    self.layout.removeWidget(cb)
                    cb.deleteLater()

            # Adds the new checkboxes to the list
            self.PTM_CBs = []
            for ptm in PTMs:
                self.PTM_CBs.append(QCheckBox(ptm, self))
            # Removes the label
            i = 0
            if not self.layout.itemAtPosition(5, i) is None:
                label_to_remove = self.layout.itemAtPosition(5, i).widget()
                self.layout.removeWidget(label_to_remove)
                label_to_remove.deleteLater()
            
            # Inserts the new PTM checkboxes
            self.layout.addWidget(QLabel("PTMs for Current Analysis"), 5, i)
            for widget in self.PTM_CBs:
                self.layout.addWidget(widget, 6, i)
                i += 1
        self.statusBar().showMessage("")
    
    @pyqtSlot()
    def onClickGenericButton(self):
        self.statusBar().showMessage("Choose the Search Engine output")
        self.statusBar().setStyleSheet("background-color : white")
        filename = self.openCSVFileNameDialog()
        #if self.generic_filepath.text() != "":
        if os.path.exists(filename):
            self.generic_filepath.setText(filename)
            missed_cleavages, protease, PTMs = parse_output("generic", filename)
            if missed_cleavages == -3:
                self.statusBar().showMessage("A sequence in the Peptide Modified Sequence column is missing PTMs")
                self.statusBar().setStyleSheet("background-color : red")
                return
            if missed_cleavages == -4:
                self.statusBar().showMessage("Invalid file")
                self.statusBar().setStyleSheet("background-color : red")
                self.generic_filepath.setText("")
                return
            #Remove the previous PTM checkboxes (if there were any)
            if self.PTM_CBs != []:
                for cb in self.PTM_CBs:
                    self.layout.removeWidget(cb)
                    cb.deleteLater()

            # Adds the new checkboxes to the list
            self.PTM_CBs = []
            for ptm in PTMs:
                self.PTM_CBs.append(QCheckBox(ptm, self))
            # Removes the label# This is where the MQ localization probability threshold will be
            i = 0
            if not self.layout.itemAtPosition(5, i) is None:
                label_to_remove = self.layout.itemAtPosition(5, i).widget()
                self.layout.removeWidget(label_to_remove)
                label_to_remove.deleteLater()

            # Inserts the new PTM checkboxes
            self.layout.addWidget(QLabel("PTMs for Current Analysis"), 5, i)
            for widget in self.PTM_CBs:
                self.layout.addWidget(widget, 6, i)
                i += 1
        self.statusBar().showMessage("")

    @pyqtSlot()
    def on_click_proteome_button(self):
        self.statusBar().showMessage("Choose the Uniprot Proteome FASTA file")
        self.statusBar().setStyleSheet("background-color : white")
        filename = self.openFASTAFileNameDialog()
        if os.path.exists(filename):
            # Parse the file to check it is valid and grab the species
            was_error, species_ID = parse_proteome(filename)
            if was_error:
                self.statusBar().showMessage("ERROR: Invalid proteome FASTA file!")
                self.statusBar().setStyleSheet("background-color : red")
            else:
                self.proteome_filepath.setText(filename)
                if not species_ID is None:
                    self.species_id_edit.setText(species_ID)
                    self.update_species_name()
        self.statusBar().showMessage("")

    @pyqtSlot()
    def on_click_output_button(self):
        self.statusBar().showMessage("Choose the ProteoSushi output file location")
        self.statusBar().setStyleSheet("background-color : white")
        filename = self.write_output_dialog()
        self.output_filepath.setText(filename)
        self.statusBar().showMessage("")

    @pyqtSlot()
    def onClickTargetButton(self):
        self.statusBar().showMessage("Choose the Target Gene file")
        self.statusBar().setStyleSheet("background-color : white")
        filename = self.openTXTFileNameDialog()
        if self.target_filepath.text() != "":
            self.target_filepath.setText(filename)
        self.statusBar().showMessage("")

    def __run_single_thread(self, fdr: float, combine_method, species_id, localization: float):
        """Runs the background on the same thread if there are problems"""
        # If the maxquant option was chosen, it sends that info to be run
        if self.maxquant_RB.isChecked() and os.path.exists(self.maxquant_filepath.text()):
            start_time = time()
            self.statusBar().showMessage("Analysis in Progress")
            self.statusBar().setStyleSheet("background-color : white")
            output = rollup("maxquant", 
                            self.maxquant_filepath.text(), 
                            self.target_checkbox.isChecked(),  # Whether target will be used
                            self.target_filepath.text(),
                            int(self.max_missed_edit.text()),
                            self.protease_combo_box.currentText(),
                            fdr,
                            self.quant_CB.isChecked(),
                            self.convert_PTM_list(self.PTM_CBs),
                            self.proteome_filepath.text(),
                            combine_method,
                            self.uniprot_annot_CB.isChecked(),
                            species_id,
                            self.output_filepath.text(),
                            localization)
            # If there is a 502 proxy error (server side error)
            if self.uniprot_annot_CB.isChecked() and output == 502:
                self.statusBar().showMessage("ERROR: Uniprot server error! Please try again later.")
                self.statusBar().setStyleSheet("background-color : red")
                return
            self.statusBar().showMessage("Analysis Complete!")
            self.statusBar().setStyleSheet("background-color : green")
            print("\033[92m {}\033[00m".format("\nAnalysis Complete!"))
            print("\033[96m {}\033[00m".format(f"Rollup took {time() - start_time} seconds"))
            #sys.exit()
        elif self.mascot_RB.isChecked() and os.path.exists(self.mascot_filepath.text()):
            start_time = time()
            self.statusBar().showMessage("Analysis in Progress")
            self.statusBar().setStyleSheet("background-color : white")
            output = rollup("mascot", 
                            self.mascot_filepath.text(), 
                            self.target_checkbox.isChecked(),  # Whether target will be used
                            self.target_filepath.text(),
                            int(self.max_missed_edit.text()),
                            self.protease_combo_box.currentText(),
                            fdr,
                            self.quant_CB.isChecked(),
                            self.convert_PTM_list(self.PTM_CBs),
                            self.proteome_filepath.text(),
                            combine_method,
                            self.uniprot_annot_CB.isChecked(),
                            species_id,
                            self.output_filepath.text(), 
                            None)
            # If the mascot file has no intensity values and user tried to analyze them
            if self.quant_CB.isChecked() and output == 2:
                self.statusBar().showMessage("ERROR: Mascot file has no detectable intensity values!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            # If there is a 502 proxy error (server side error)
            if self.uniprot_annot_CB.isChecked() and output == 502:
                self.statusBar().showMessage("ERROR: Uniprot server error! Please try again later.")
                self.statusBar().setStyleSheet("background-color : red")
                return
            self.statusBar().showMessage("Analysis Complete!")
            self.statusBar().setStyleSheet("background-color : green")
            print("\033[92m {}\033[00m".format("\nAnalysis Complete!"))
            print("\033[96m {}\033[00m".format(f"Rollup took {time() - start_time} seconds"))
            #sys.exit()
        elif self.generic_RB.isChecked() and os.path.exists(self.generic_filepath.text()):
            start_time = time()
            self.statusBar().showMessage("Analysis in Progress")
            self.statusBar().setStyleSheet("background-color : white")
            output = rollup("generic", 
                            self.generic_filepath.text(), 
                            self.target_checkbox.isChecked(),  # Whether target will be used
                            self.target_filepath.text(),
                            int(self.max_missed_edit.text()),
                            self.protease_combo_box.currentText(),
                            fdr,
                            self.quant_CB.isChecked(),
                            self.convert_PTM_list(self.PTM_CBs),
                            self.proteome_filepath.text(),
                            combine_method,
                            self.uniprot_annot_CB.isChecked(),
                            species_id,
                            self.output_filepath.text(),
                            localization)
            # If there is a 502 proxy error (server side error)
            if self.uniprot_annot_CB.isChecked() and (output == 502 or output == 4):
                self.statusBar().showMessage("ERROR: Uniprot server error! Please try again later.")
                self.statusBar().setStyleSheet("background-color : red")
                return
            # If the generic file has no intensity values and user tried to analyze them
            if self.quant_CB.isChecked() and output == 2:
                self.statusBar().showMessage("ERROR: Generic file has no detectable intensity values!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            self.statusBar().showMessage("Analysis Complete!")
            self.statusBar().setStyleSheet("background-color : green")
            print("\033[92m {}\033[00m".format("\nAnalysis Complete!"))
            print("\033[96m {}\033[00m".format(f"Rollup took {time() - start_time} seconds"))
            #sys.exit()
        else:
            self.statusBar().showMessage("ERROR: Missing Search Engine Output!")
            self.statusBar().setStyleSheet("background-color : red")


    def __run_async(self, fdr: float, combine_method, species_id, localization: float):
        """Function that runs the backend on a separate thread"""
        # If the maxquant option was chosen, it sends that info to be run
        if self.maxquant_RB.isChecked() and os.path.exists(self.maxquant_filepath.text()):
            start_time = time()
            self.statusBar().showMessage("Analysis in Progress")
            self.statusBar().setStyleSheet("background-color : white")
            output = rollup("maxquant", 
                            self.maxquant_filepath.text(), 
                            self.target_checkbox.isChecked(),  # Whether target will be used
                            self.target_filepath.text(),
                            int(self.max_missed_edit.text()),
                            self.protease_combo_box.currentText(),
                            fdr,
                            self.quant_CB.isChecked(),
                            self.convert_PTM_list(self.PTM_CBs),
                            self.proteome_filepath.text(),
                            combine_method,
                            self.uniprot_annot_CB.isChecked(),
                            species_id,
                            self.output_filepath.text(),
                            localization)
            # If there is a 502 proxy error (server side error)
            if self.uniprot_annot_CB.isChecked() and output == 502:
                self.statusBar().showMessage("ERROR: Uniprot server error! Please try again later.")
                self.statusBar().setStyleSheet("background-color : red")
                return
            self.statusBar().showMessage("Analysis Complete!")
            self.statusBar().setStyleSheet("background-color : green")
            print("\033[92m {}\033[00m".format("\nAnalysis Complete!"))
            print("\033[96m {}\033[00m".format(f"Rollup took {time() - start_time} seconds"))
            #sys.exit()
        elif self.mascot_RB.isChecked() and os.path.exists(self.mascot_filepath.text()):
            start_time = time()
            self.statusBar().showMessage("Analysis in Progress")
            self.statusBar().setStyleSheet("background-color : white")
            output = rollup("mascot", 
                            self.mascot_filepath.text(), 
                            self.target_checkbox.isChecked(),  # Whether target will be used
                            self.target_filepath.text(),
                            int(self.max_missed_edit.text()),
                            self.protease_combo_box.currentText(),
                            fdr,
                            self.quant_CB.isChecked(),
                            self.convert_PTM_list(self.PTM_CBs),
                            self.proteome_filepath.text(),
                            combine_method,
                            self.uniprot_annot_CB.isChecked(),
                            species_id,
                            self.output_filepath.text(), 
                            None)
            # If the mascot file has no intensity values and user tried to analyze them
            if self.quant_CB.isChecked() and output == 2:
                self.statusBar().showMessage("ERROR: Mascot file has no detectable intensity values!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            # If there is a 502 proxy error (server side error)
            if self.uniprot_annot_CB.isChecked() and output == 502:
                self.statusBar().showMessage("ERROR: Uniprot server error! Please try again later.")
                self.statusBar().setStyleSheet("background-color : red")
                return
            self.statusBar().showMessage("Analysis Complete!")
            self.statusBar().setStyleSheet("background-color : green")
            print("\033[92m {}\033[00m".format("\nAnalysis Complete!"))
            print("\033[96m {}\033[00m".format(f"Rollup took {time() - start_time} seconds"))
            #sys.exit()
        elif self.generic_RB.isChecked() and os.path.exists(self.generic_filepath.text()):
            start_time = time()
            self.statusBar().showMessage("Analysis in Progress")
            self.statusBar().setStyleSheet("background-color : white")
            output = rollup("generic", 
                            self.generic_filepath.text(), 
                            self.target_checkbox.isChecked(),  # Whether target will be used
                            self.target_filepath.text(),
                            int(self.max_missed_edit.text()),
                            self.protease_combo_box.currentText(),
                            fdr,
                            self.quant_CB.isChecked(),
                            self.convert_PTM_list(self.PTM_CBs),
                            self.proteome_filepath.text(),
                            combine_method,
                            self.uniprot_annot_CB.isChecked(),
                            species_id,
                            self.output_filepath.text(),
                            localization)
            # If there is a 502 proxy error (server side error)
            if self.uniprot_annot_CB.isChecked() and (output == 502 or output == 4):
                self.statusBar().showMessage("ERROR: Uniprot server error! Please try again later.")
                self.statusBar().setStyleSheet("background-color : red")
                return
            # If the generic file has no intensity values and user tried to analyze them
            if self.quant_CB.isChecked() and output == 2:
                self.statusBar().showMessage("ERROR: Generic file has no detectable intensity values!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            self.statusBar().showMessage("Analysis Complete!")
            self.statusBar().setStyleSheet("background-color : green")
            print("\033[92m {}\033[00m".format("\nAnalysis Complete!"))
            print("\033[96m {}\033[00m".format(f"Rollup took {time() - start_time} seconds"))
            #sys.exit()
        else:
            self.statusBar().showMessage("ERROR: Missing Search Engine Output!")
            self.statusBar().setStyleSheet("background-color : red")
        #sys.exit()

    def __cut_thread(self):
        """Called once the backend thread finishes"""
        # Enables the buttons again in case I get multiple analyses working
        self.run_button.setEnabled(True)
        self.proteome_filepath_button.setEnabled(True)
        self.output_filepath_button.setEnabled(True)
        self.maxquant_button.setEnabled(True)
        self.mascot_button.setEnabled(True)
        self.generic_button.setEnabled(True)
        #sys.exit()
        return

    @pyqtSlot()
    def onClickRunButton(self):
        if ((self.maxquant_RB.isChecked() and os.path.exists(self.maxquant_filepath.text())) or
            (self.mascot_RB.isChecked() and os.path.exists(self.mascot_filepath.text())) or
            (self.generic_RB.isChecked() and os.path.exists(self.generic_filepath.text()))): 
            if not os.path.exists(self.proteome_filepath.text()):
                self.statusBar().showMessage("ERROR: Missing Proteome FASTA file!")
                self.statusBar().setStyleSheet("background-color : red")
                return

            # If the user didn't choose a name or location for the file
            if self.output_filepath.text() == "":
                self.statusBar().showMessage("ERROR: ProteoSushi output name and location not chosen!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            
            if os.path.isdir(self.output_filepath.text()):
                self.output_filepath.setText(os.path.join(self.output_filepath.text(), "proteosushi_output.csv"))

            if self.output_filepath.text() == "[Filepath]":
                self.output_filepath.setText(os.path.join(os.getcwd(), "proteosushi_output.csv"))

            if self.output_filepath.text()[-4:] != ".csv":
                self.output_filepath.setText(self.output_filepath.text() + ".csv")

            try:  # Check to see if the provided number of allowed missed cleavages is legal
                max_missed = int(self.max_missed_edit.text())
                if max_missed < 0:
                    self.statusBar().showMessage("ERROR: Number of max allowed missed cleavages cannot be less than 0!")
                    self.statusBar().setStyleSheet("background-color : red")
                    return
            except ValueError:
                self.statusBar().showMessage("ERROR: Invalid number for max allowed missed cleavages!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            # Checks that the localization prob threshold is between 0 and 1
            try:
                if self.localization_edit.text() != "":
                    localization = float(self.localization_edit.text())
                    if localization < 0 or localization > 1:
                        self.statusBar().showMessage("ERROR: Localization threshold must be between 0 and 1!")
                        self.statusBar().setStyleSheet("background-color : red")
                        return
                else:
                    localization = None
            except ValueError:
                self.statusBar().showMessage("ERROR: Localization threshold must be a number between 0 and 1!")
                self.statusBar().setStyleSheet("background-color : red")
                return

            try:  # Checks to see if the provided FDR threshold is legal
                if self.fdr_edit.text() != "":
                    fdr = float(self.fdr_edit.text())
                    if fdr < 0 or fdr > 1:
                        self.statusBar().showMessage("ERROR: FDR must be between 0 and 1!")
                        self.statusBar().setStyleSheet("background-color : red")
                        return
                else:
                    fdr = None
            except ValueError:
                self.statusBar().showMessage("ERROR: FDR must be a number")
                self.statusBar().setStyleSheet("background-color : red")
                return
            # If quantitation will be used, but the user has not selected a method to combine
            if self.quant_CB.isChecked() and not self.sum_RB.isChecked() and not self.average_RB.isChecked():
                self.statusBar().showMessage("ERROR: Must choose either Sum or Average for the quant method!")
                self.statusBar().setStyleSheet("background-color : red")
                return
            elif self.quant_CB.isChecked() and self.sum_RB.isChecked():
                combine_method = "sum"
            elif self.quant_CB.isChecked() and self.average_RB.isChecked():
                combine_method = "average"
            else:
                combine_method = ""
            
            # If a valid species ID has been entered in
            species_id = self.species_id_edit.text()
            if len(species_id) > 0 and not species_id in self.species_id_dict:
                self.statusBar().showMessage("ERROR: Not a valid species ID!")
                self.statusBar().setStyleSheet("background-color : red")
                return

                
            if self.protease_combo_box.currentText() in cleave_rules:
                # Disable all buttons before running to keep analysis confined
                self.run_button.setEnabled(False)
                self.proteome_filepath_button.setEnabled(False)
                self.output_filepath_button.setEnabled(False)
                self.maxquant_button.setEnabled(False)
                self.mascot_button.setEnabled(False)
                self.generic_button.setEnabled(False)

                if not self.multithread:
                    self.__run_single_thread(fdr, combine_method, species_id, localization)
                else:
                    # Starts a new thread for the backend analysis; causes seg fault currently
                    print("\033[96m {}\033[00m".format("\nAnalysis Started!"))
                    worker = Worker(self.__run_async, fdr, combine_method, species_id, localization)
                    worker.signals.end.connect(self.__cut_thread)
                    self.threadpool.start(worker)
                
            else:
                self.statusBar().showMessage("ERROR: Protease provided is not supported!")
                self.statusBar().setStyleSheet("background-color : red")
        else:
            self.statusBar().showMessage("ERROR: Missing Search Engine Output!")
            self.statusBar().setStyleSheet("background-color : red")




    def createGridLayout(self):
        self.horizontalGroupBox = QGroupBox("")
        self.layout = QGridLayout()
        self.layout.setRowStretch(0, 1)
        self.layout.setRowStretch(9, 1)

        row = 1
        self.layout.addWidget(self.file_chooser_label, row, 0)
        row += 1
        self.layout.addWidget(self.maxquant_RB, row, 0)
        self.layout.addWidget(self.maxquant_button, row, 1)
        self.layout.addWidget(self.maxquant_filepath, row, 2)
        row += 1
        self.layout.addWidget(self.mascot_RB, row, 0)
        self.layout.addWidget(self.mascot_button, row, 1)
        self.layout.addWidget(self.mascot_filepath, row, 2)
        row += 1
        self.layout.addWidget(self.generic_RB, row, 0)
        self.layout.addWidget(self.generic_button, row, 1)
        self.layout.addWidget(self.generic_filepath, row, 2)
        row += 1
        # This is where the PTM label goes
        row += 1
        # This is where the PTM checkboxes go
        row += 1
        self.layout.addWidget(QLabel("Proteome File", self), row, 0)
        self.layout.addWidget(self.proteome_filepath_button, row, 1)
        self.layout.addWidget(self.proteome_filepath, row, 2)
        row += 1
        self.layout.addWidget(self.species_id_label, row, 0)
        self.layout.addWidget(self.species_id_edit, row, 1)
        self.layout.addWidget(self.species_name_label, row, 2)
        row += 1
        self.layout.addWidget(self.max_missed_label, row, 0)
        self.layout.addWidget(self.max_missed_edit, row, 1)
        row += 1
        self.layout.addWidget(self.protease_label, row, 0)
        self.layout.addWidget(self.protease_combo_box, row, 1)
        row += 1
        self.layout.addWidget(QLabel("Output Name & Location", self), row, 0)
        self.layout.addWidget(self.output_filepath_button, row, 1)
        self.layout.addWidget(self.output_filepath, row, 2)
        row += 1
        self.layout.addWidget(self.options_label, row, 0)
        row += 1
        self.layout.addWidget(self.localization_checkbox, row, 0)
        self.layout.addWidget(self.localization_edit, row, 1)
        row += 1
        self.layout.addWidget(self.target_checkbox, row, 0)
        self.layout.addWidget(self.target_button, row, 1)
        self.layout.addWidget(self.target_filepath, row, 2)
        row += 1
        self.layout.addWidget(self.quant_CB, row, 0)
        self.layout.addWidget(self.sum_RB, row, 1)
        self.layout.addWidget(self.average_RB, row, 2)
        row += 1
        self.layout.addWidget(self.uniprot_annot_CB, row, 0)
        row += 1
        self.layout.addWidget(self.fdr_CB, row, 0)
        self.layout.addWidget(self.fdr_edit, row, 1)
        row += 1
        self.layout.addWidget(self.run_button, row, 0)

        self.layout.setAlignment(Qt.AlignTop)
        self.horizontalGroupBox.setAlignment(Qt.AlignTop)

        self.horizontalGroupBox.setLayout(self.layout)


def run_gui(multithread: bool):
    app = QApplication(sys.argv)
    ex = App(multithread)
    sys.exit(app.exec_())

if __name__ == "__main__":
    app = QApplication(sys.argv)
    ex = App()
    sys.exit(app.exec_())



"""parse_mascot.py: functions used to parse the mascot file and read it into memory"""

import csv
from collections import defaultdict
import os
from re import match
from time import sleep
try:
    from .ps_utilities import clean_pep_seq, parse_mascot, load_pepdict #most of the parsing actually happens here 
except ImportError:
    from ps_utilities import clean_pep_seq, parse_mascot, load_pepdict


def __create_mod_dict(filename: str, mod_ids: list, var_mod_map: dict, cleave_rule: tuple, PTMs: list) -> dict:
    """Makes the PTM dictionary that connects a peptide sequence with its modifications
    Arguments:
        filename {str} -- the name of the output file
        mod_ids {list} -- a list of mod ids that mascot uses and the user chose
        var_mod_map {dict} -- connects the PTM to the Mascot index

    Returns:
        dict -- a dictionary of sequences with the type and location of modifications
    """
    modDict = {}
    inv_mod_map = {v:k for k, v in var_mod_map.items()}
    with open(filename, 'r') as mascot_output:
        file_reader = csv.reader(mascot_output, quotechar='"')
        header = next(file_reader)
        sequence = header.index("pep_seq")
        var_mods = header.index("pep_var_mod_pos")

        for row in file_reader:
            #Mascot has duplicate rows; check for these
            pep_seq = row[sequence].replace("L","I")
            if len(pep_seq) < 6:  # Skip any peptides smaller than 6
                continue
            #print(row[sequence])
            mods = row[var_mods]
            if not mods:
                continue
            pep_mod_seq = mods.split('.')[1]
            i = len(pep_mod_seq) - 1
            inv_mod_dict = {v:k for k, v in var_mod_map.items()}
            new_seq = row[sequence]
            while i >= 0:
                if pep_mod_seq[i] != "0":  # If there is a mod
                    new_seq = row[sequence][:i+1] + '(' + inv_mod_dict[pep_mod_seq[i]] + ')' + new_seq[i+1:]
                i -= 1
            pep_mod_seq = new_seq
            new_mod_seq, new_pep_seq, missed_cleave_fix = clean_pep_seq(cleave_rule, pep_mod_seq, PTMs, row[sequence])
            if mods:
                n, center, c = [x for x in mods.split(".")]
                for i, aa in enumerate(list(center)): #split mods; note this excludes termini
                    if aa in mod_ids:
                        try:
                            if not tuple((str(aa), i)) in modDict[pep_mod_seq]:
                                modDict[new_mod_seq].append(tuple((inv_mod_map[aa], i-missed_cleave_fix)))
                        except KeyError:
                            modDict[new_mod_seq] = [tuple((inv_mod_map[aa], i-missed_cleave_fix))]
            #if row[sequence] == "AAFTECCQAADK":
            #    input(f"old {row[sequence]}, new {new_mod_seq}")
    return modDict


def compile_data_mascot(search_engine_filepath: str, PTMs: list, cleave_rule: tuple) -> list:
    """Takes the lists and dictionaries needed to parse files

    Arguments:
        search_engine_filepath {str} -- the filepath to the mascot output file
        PTMs {list} -- the list of PTMs that the user has chosen to use
    Returns:
        int -- index of unmodified sequence
        int -- index of modified sequence
        dict -- mod_dict
        int -- intensity_start 
        str -- data_filename
        dict -- dictionary of the mods to the index
    """
    in_file = search_engine_filepath
    protease, quant_range, var_mod_map, missed_cleaves = parse_mascot(in_file)

    mods_for_quant = PTMs


    input_filename = "quant_parsed.csv"
    input_file =  open(input_filename, 'r')
    file_reader = csv.reader(input_file)
    header = next(file_reader)

    sequence = header.index('pep_seq') #assert this exists later on
    var_mods = header.index('pep_var_mod_pos')

    mod_dict = __create_mod_dict(input_filename, [var_mod_map[mod] for mod in mods_for_quant], var_mod_map, cleave_rule, PTMs)
    #print("3")
    if quant_range[1] <= quant_range[0]:
        return sequence, var_mods, mod_dict, None, input_filename, var_mod_map

    return sequence, var_mods, mod_dict, quant_range[0], input_filename, var_mod_map
#EOF


"""run_proteoSushi.py: The starting point to run proteoSushi"""
import logging
import os

try:
    from proteosushi.proteosushi_gui import run_gui
    from proteosushi.combine_intensities import rollup
except ImportError:  # Allows the program to be run not as a module
    from proteosushi_gui import run_gui
    from combine_intensities import rollup


if __name__ == "__main__":
    try:
        os.mkdir("logs")
    except FileExistsError:
        pass
    logging.basicConfig(level=logging.DEBUG, filename=os.path.join("logs","ps.log"), filemode='w')
    run_gui(multithread=True)
    '''
    rollup(
        search_engine="generic",
        #search_engine_filepath="/home/rob/Documents/Held_Lab/ProteoSushi/proteosushi/examples/EGFR_Skyline_data.csv",
        search_engine_filepath="/home/rob/Downloads/Proteosushi_test_deamid_seqOnly.csv",
        use_target_list=False,
        target_list_filepath="",
        max_missed_cleavages=2,
        protease="trypsin/p",
        fdr_threshold=None,
        use_quant=False,
        #user_PTMs=["C[+57]", "C[+IAC/NEM]"],
        user_PTMs=["N[+0.984]"],
        #proteome_fasta_filepath="/home/rob/Documents/Held_Lab/ProteoSushi/proteosushi/examples/fastas/Uni-Hum-Ref-20141022.fasta",
        proteome_fasta_filepath="/home/rob/Downloads/9606_HomoSapiens_150721.fasta",
        #intensity_method="sum",
        intensity_method="",
        add_annotation=True,
        species_id="9606",
        output_filename="/home/rob/Documents/Held_Lab/ProteoSushi/proteosushi/output/ps_test_211120.csv",
        localization_threshold=None
    )
    '''
#EOF


#!/usr/bin/env python

"""parse_MaxQuant.py: using helper functions, the script parses MaxQuant result files"""

from collections import defaultdict
import csv
import os.path
from re import findall, finditer, match
from time import sleep
try:
    from .ps_utilities import clean_localization_pep_seq, clean_pep_seq
    from .ps_utilities import load_pepdict, parse_maxquant_summary
except ImportError:
    from ps_utilities import clean_localization_pep_seq, clean_pep_seq
    from ps_utilities import load_pepdict, parse_maxquant_summary


def get_MQ_PTMs(filename: str) -> list:
    """Parses evidence file for just the PTMs available for selection
    Arguments:
        filename {str} -- name of the evidence file with the location
    Returns:
        list -- a list of the PTMs we can use for the analysis
    """
    PTMs = list()
    with open(filename, 'r') as evidence_file:
        tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
        header = next(tsv_reader)
        header_lower = [s.lower() for s in header]
        start_index = header_lower.index([ptm for ptm in header_lower if "score diffs" in ptm][-1]) + 1
        end_index = header_lower.index("missed cleavages")
        PTMs = header[start_index:end_index]
    return PTMs


def parse_evidence_localization(filename: str, user_PTMs: list, cleave_rule: tuple, localization_threshold: float) -> dict:
    """parses the evidence file and builds a dictionary for PTMs
    Arguments:
        filename {str} -- name of the evidence file with the location
        user_PTMs {list} -- list of user-chosen PTMs for analysis
        cleave_rule {tuple} -- information on the protease for cleaving in silico
        localization_threshold {float} -- the user-chosen threshold for localization probability
    Returns:
        dict -- a dictionary of the PTMs with sequence as key and a list of tuples with 
        PTM/location as value
        list -- a list of the PTMs we can use for the analysis
    """
    mod_dict = dict()
    PTMs = list()
    with open(filename, 'r') as evidence_file:
        tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
        header = next(tsv_reader)
        header_lower = [s.lower() for s in header]
        seq_index = header_lower.index("sequence")
        mod_indices = [i for i, s in enumerate(header) if "Probabilities" in s]
        modified_index = header_lower.index("modified sequence")
        PTMs = [ptm.split(" Probabilities")[0] for ptm in header if "Probabilities" in ptm]
        # Go through each line of the evidence file
        for row in tsv_reader:
            mod_seq = row[modified_index].strip('_')
            if len(row[seq_index]) < 6:
                continue
            loc_seqs = [row[i] if header[i].split(" Probabilities")[0] in user_PTMs else "" for i in mod_indices]
            new_mod_seq, new_pep_seq, missed_cleave_fix = clean_pep_seq(cleave_rule, 
                                                                        mod_seq, 
                                                                        user_PTMs, 
                                                                        row[seq_index],
                                                                        True)
            new_loc_seqs, new_pep_seq, missed_cleave_fix = clean_localization_pep_seq(cleave_rule, 
                                                                                      loc_seqs, 
                                                                                      user_PTMs, 
                                                                                      row[seq_index])
            
            if new_loc_seqs is None:
                continue
            # Go through each of the PTM localization sequences
            index = 0
            while index < len(new_loc_seqs):
                if new_loc_seqs[index] == "":
                    index += 1
                    continue
                loc_iter = finditer(r"(\[.+?\])|(\(.+?\))", new_loc_seqs[index])
                loc_indices = []
                for loc_i in loc_iter:
                    loc_indices.append(loc_i.start())
                
                try:
                    loc_probs = findall(r"(\[.+?\])|(\(.+?\))", new_loc_seqs[index]).remove('')
                except ValueError:
                    loc_probs = findall(r"(\[.+?\])|(\(.+?\))", new_loc_seqs[index])
                
                index_correction = 0
                j = 0
                # Go through each of the PTM sites
                while j < len(loc_probs):
                    # Skips sites below the threshold
                    if float(loc_probs[j][1].replace('(','').replace(')','').replace('[','').replace(']','')) < localization_threshold:
                        index_correction += len(loc_probs[j][1])
                        j += 1
                        continue
                    try:  # The - 1 is a hotfix as I'm not sure why it is just 1 off.
                        mod_dict[new_mod_seq].append(tuple((PTMs[index], loc_indices[j] - index_correction - 1)))
                        index_correction += len(loc_probs[j][1])
                        j += 1
                    except KeyError:
                        mod_dict[new_mod_seq] = [tuple((PTMs[index], loc_indices[j] - index_correction - 1))]
                        index_correction += len(loc_probs[j][1])
                        j += 1
                index += 1
    return mod_dict, PTMs

def parse_evidence(filename: str, user_PTMs: list, cleave_rule: tuple) -> dict:
    """parses the evidence file and builds a dictionary for PTMs
    Arguments:
        filename {str} -- name of the evidence file with the location
    Returns:
        dict -- a dictionary of the PTMs with sequence as key and a list of tuples with 
        PTM/location as value
        list -- a list of the PTMs we can use for the analysis
    """
    mod_dict = dict()
    PTMs = list()
    with open(filename, 'r') as evidence_file:
        tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
        header = next(tsv_reader)
        header_lower = [s.lower() for s in header]
        seq_index = header_lower.index("sequence")
        modified_index = header_lower.index("modified sequence")
        mods_index = header_lower.index("modifications")
        start_index = header_lower.index([ptm for ptm in header_lower if "score diffs" in ptm][-1]) + 1
        end_index = header_lower.index("missed cleavages")
        PTMs = header[start_index:end_index]
        for row in tsv_reader:
            pep_mods = dict()
            modified_sequence = row[modified_index].strip('_')
            new_mod_seq, new_pep_seq, missed_cleave_fix = clean_pep_seq(cleave_rule, modified_sequence, user_PTMs, row[seq_index])
            abr_ptms = [PTM[:2].lower() for PTM in user_PTMs]

            mods = findall(r"(\[.+?\])|(\(.+?\))", new_mod_seq)
            breaks = finditer(r"(\[.+?\])|(\(.+?\))", new_mod_seq)
            cut_sites = []
            for breakp in breaks:
                cut_sites.append(breakp.start())
            # This fixes the indices of each mod to connect with the unmodified sequence
            correction = 0
            fixed_indices = []
            num_of_mods = 0
            for site in cut_sites:
                fixed_indices.append(site - correction - 1)  # TODO: Check this!
                correction += len(mods[num_of_mods][1])
                num_of_mods += 1

            #if row[seq_index] == "AASCVLLHTGQK":
            #    input(str(mods))
            i = 0
            while i < len(cut_sites):
                if not mods[i][1][1:-1] in abr_ptms:
                    #print(f"{mods[i][1][1:-1]} not in {abr_ptms}")
                    i += 1
                    continue
                try:
                    mod_dict[new_mod_seq].append(tuple(([ptm for ptm in user_PTMs if mods[i][1][1:-1] == ptm[:2].lower()][0], 
                                                        fixed_indices[i])))
                except KeyError:
                    mod_dict[new_mod_seq] = [tuple(([ptm for ptm in user_PTMs if mods[i][1][1:-1] == ptm[:2].lower()][0], 
                                                    fixed_indices[i]))]
                i += 1

            '''
            for PTM in PTMs:
                modified_sequence = row[modified_index].strip('_')
                new_mod_seq, new_pep_seq, missed_cleave_fix = clean_pep_seq(cleave_rule, modified_sequence, user_PTMs, row[seq_index])
                mods_B4_mod = 0
                # Grabs the first two letters of PTM that maxquant uses in mod_seq
                abr_ptm = PTM[:2].lower()  # NOTE: I could possibly imagine this not working in certain circumstances...
                # while this PTM is in the sequence we are looking at and listed among the PTMs for this peptide
                while PTM in row[mods_index] and abr_ptm in new_mod_seq:  # This seems a little sketchy. What if there are two different PTMs with the same first 2 letters?
                    other_PTMs = PTMs.copy()
                    other_PTMs.remove(PTM)
                    # Gets the index of the PTM within the sequence (that is right here XX(|ptm)XX)
                    for other_ptm in other_PTMs:
                        mod_index = new_mod_seq.index(abr_ptm)
                        abr_other_ptm = other_ptm[:2].lower()
                        while (abr_other_ptm in new_mod_seq and 
                               new_mod_seq.index(abr_other_ptm) < mod_index):
                            #This takes the ptms, one by one, until each before is removed
                            mods_B4_mod += 1
                            ptm_index = new_mod_seq.index(abr_other_ptm)
                            new_mod_seq = (new_mod_seq[:ptm_index]
                                               + new_mod_seq[ptm_index + len(abr_other_ptm):])  # NOTE: removes the mod, but not the parenthesis
                    mod_index = new_mod_seq.index(abr_ptm)
                    pep_mods[new_mod_seq.index(abr_ptm) - (mods_B4_mod * 2)] = PTM
                    new_mod_seq = (new_mod_seq[:mod_index]
                                       + new_mod_seq[mod_index + len(abr_ptm):])
                    mods_B4_mod += 1
            
            for loc in sorted(pep_mods):
                try:
                    if not tuple((pep_mods[loc], loc - 2)) in mod_dict[new_mod_seq]:
                        mod_dict[new_mod_seq].append(tuple((pep_mods[loc],
                                                     loc - 2 - missed_cleave_fix)))
                except KeyError:
                    mod_dict[new_mod_seq] = [tuple((pep_mods[loc],
                                             loc - 2 - missed_cleave_fix))]
            '''
    return mod_dict, PTMs

def compile_localization_data_maxquant(search_engine_filepath: str, user_PTMs: list, 
                                       cleave_rule: tuple, localization_threshold: float) -> list:
    """Takes the lists and dictionaries needed to parse files
    Arguments:
        search_engine_filepath {str} -- filepath to the SE directory
        user_PTMs {list} -- a list of user-selected PTMs for analysis
        cleave_rule {tuple} -- the protease cleave rule
        localization_threshold {float} -- user-chosen threshold for localization probability
    Returns:
        int -- index of unmodified sequence
        int -- index of modified sequence
        list -- list of indices of modified sequence w/ localization prob
        dict -- mod_dict
        int -- intensity_start (maybe just skyline)
        str -- evidence_filename
    """
    MQ_dir = search_engine_filepath
    sum_file = os.path.join(MQ_dir, "summary.txt")
    protease, max_missed_cleaves = parse_maxquant_summary(sum_file)
    evidence_filename = os.path.join(MQ_dir, "evidence.txt")
    mod_dict, PTMs = parse_evidence_localization(evidence_filename, user_PTMs, cleave_rule, localization_threshold)  # This grabs the PTMs from the evidence file
    PTMs = user_PTMs
    evidence_file = open(evidence_filename, 'r')
    tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
    header = next(tsv_reader)
    header_lower = [s.lower() for s in header]
    sequence = header_lower.index("sequence")
    mod_seq = header_lower.index("modified sequence")
    loc_seqs = [i for i, h in enumerate(header) if " probabilities" in h.lower()]
    reporter_intensities = [i for i, h in enumerate(header) if "intensit" in h.lower() and 
                                                               not "max intensity m/z" in h.lower()]
    evidence_file.close()
    return sequence, mod_seq, loc_seqs, mod_dict, reporter_intensities, evidence_filename

def compile_data_maxquant(search_engine_filepath: str, user_PTMs: list, cleave_rule: tuple) -> list:
    """Takes the lists and dictionaries needed to parse files
    Arguments:
        search_engine_filepath {str} -- filepath to the SE directory
        user_PTMs {list} -- a list of user-selected PTMs for analysis
        cleave_rule {tuple} -- the protease cleave rule
    Returns:
        int -- index of unmodified sequence
        int -- index of modified sequence
        dict -- mod_dict
        int -- intensity_start (maybe just skyline)
        str -- evidence_filename
        list -- PTMs the user has chosen to be used
    """
    MQ_dir = search_engine_filepath
    sum_file = os.path.join(MQ_dir, "summary.txt")
    protease, max_missed_cleaves = parse_maxquant_summary(sum_file)
    evidence_filename = os.path.join(MQ_dir, "evidence.txt")
    mod_dict, PTMs = parse_evidence(evidence_filename, user_PTMs, cleave_rule)  # This grabs the PTMs from the evidence file
    PTMs = user_PTMs
    evidence_file = open(evidence_filename, 'r')
    tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
    header = next(tsv_reader)
    header_lower = [s.lower() for s in header]
    sequence = header_lower.index("sequence")
    mod_seq = header_lower.index("modified sequence")
    reporter_intensities = [i for i, h in enumerate(header) if "intensit" in h.lower() and 
                                                               not "max intensity m/z" in h.lower()]
    evidence_file.close()
    return sequence, mod_seq, mod_dict, reporter_intensities, evidence_filename, PTMs
#EOF


"""sparql.py: takes the results from proteoSushi and annotates it using the sparql API to access uniprot"""

__author__ = "Rob Seymour, Arshag Mooradian"
__email__ = "rseymour@wustl.edu"

from io import StringIO
import json
import logging

import csv
from datetime import datetime
import pandas as pd
from re import findall
import requests
import sys  # Remove later
from time import sleep
import urllib


# Global constants
ENDPOINT = "https://sparql.uniprot.org"

def sparql_json_to_df(sparql_json_dict: dict):
    """Converts the messy json format into a dataframe

    Arguments:
        sparql_json_dict {dict} -- a json object retrieved through requests.get query
    Returns:
        sparql_df {pd.DataFrame} -- dataframe containing the results
    """
    
    cols = sparql_json_dict["head"]["vars"]

    temp_row_list = [] # list holding rows
    for row in sparql_json_dict["results"]["bindings"]:
        item = []
        for c in cols:
            item.append(row.get(c, {}).get("value"))
        temp_row_list.append(item)

    sparql_df = pd.DataFrame(temp_row_list, columns = cols)
    return sparql_df


def sparql_request(unpid_site_list: list):
    """Sends the request to uniprot to get data on each of the sites
    Arguments:
        unpidSiteList {list} -- a list of tuples with uniprot ID and site of PTM
    Returns:
        pandas.DataFrame -- the complete annotated dataframe to be applied to the rolled-up PTM sites
    """
    unpid_site_list_str = ""
    if not unpid_site_list:
        return None
    for tup in unpid_site_list:
        if len(tup) >= 2 and tup[1] == '0':
            tup[1] = '1'
        if len(tup[0]) >= 5:
            unpid_site_list_str += f"(uniprotkb:{tup[0]} {tup[1]})\n"
    logging.debug(f"Formatted Input String: {unpid_site_list_str}")


    prefix="""PREFIX uniprotkb: <http://purl.uniprot.org/uniprot/>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX faldo: <http://biohackathon.org/resource/faldo#>
PREFIX up: <http://purl.uniprot.org/core/>
PREFIX ec: <http://purl.uniprot.org/enzyme/>"""

    #TODO: Add up:Similarity_Annotation, 
    alt_query_entry_len_pos = prefix + """
SELECT
    ?entry
    ?position
    #?begin
    #?end
    (STRLEN(?iupac) AS ?lengthOfSequence)
    #(SUBSTR(?iupac, ?begin, ?end - ?begin + 1) AS ?regionOfInterest)
WHERE {
    {
        SELECT ?entry ?position ?sequence ?begin ?end
        WHERE {
            VALUES (?entry ?position) {""" + unpid_site_list_str + """}
            GRAPH <http://sparql.uniprot.org/uniprot> {
                ?entry up:sequence ?sequence .
                ?range1 faldo:begin  ?bp .
                ?bp faldo:position ?begin ;
                    faldo:reference ?sequence .
                    FILTER (?begin <= ?position) .
                ?range1 faldo:end ?ep .
                ?ep  faldo:position ?end ; faldo:reference ?sequence .
                    FILTER (?position <= ?end) .
            }
        }
    }
    GRAPH <http://sparql.uniprot.org/uniprot> {
        ?sequence rdf:value ?iupac .
        # ?annotation up:range ?range1;
        # a ?type .
        # ?annotation value does not seem to be used at all,
        # is this now retrieved in a different query?
    }
}"""


    query_entry_subcellular = prefix + """
SELECT
    ?entry
    ?location
WHERE {
    GRAPH <http://sparql.uniprot.org/uniprot> {
        VALUES (?entry ?position) {""" + unpid_site_list_str + """}
        ?entry up:annotation ?subAnnotation .
        ?subAnnotation a up:Subcellular_Location_Annotation .
        ?subAnnotation up:locatedIn/up:cellularComponent ?location .
    }
}"""




    query_ec_rhea_type = prefix + """
SELECT
    ?entry
    ?position
    ?ec
    ?rhea
WHERE {
    GRAPH <http://sparql.uniprot.org/uniprot> {
        VALUES (?entry ?position) {""" + unpid_site_list_str + """}
        ?entry up:sequence ?sequence .
        ?entry up:annotation ?caAnnotation .
        ?caAnnotation up:catalyticActivity ?ca .

        OPTIONAL {
            ?ca up:enzymeClass ?ec .
        }

        OPTIONAL {
            ?ca up:catalyzedReaction ?rhea .
        }
    }
}"""  # NOTE: remove the ordering to help reduce time for query

    query_comment = prefix + """
SELECT
    ?entry
    ?position
    ?type
    ?comment
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
    VALUES (?entry ?position) {""" + unpid_site_list_str + """}
    ?entry up:sequence ?sequence .
    ?entry up:annotation ?annotation .
    ?annotation up:range ?range ;
        a ?type .
    ?range faldo:begin
        [ faldo:position ?begin ; faldo:reference ?sequence ] ;
            faldo:end
        [ faldo:position ?end ; faldo:reference ?sequence ] .
    FILTER (?begin <= ?position && ?position <= ?end)
	OPTIONAL {
        ?annotation rdfs:comment ?comment .
    }
}"""

    #TODO: Integrate the new fields in the parsing
    alt_query_comment = prefix + """
SELECT
    ?entry
    ?position
    ?type
    ?comment
    ?begin 
    ?end
    ?regionOfInterest
WHERE {
    {
        # Force the query to always first find specific annotations
        # before finding out the type of the annotation and the comment .
        SELECT 
            ?entry 
            ?position 
            ?begin
            ?end
            (SUBSTR(?iupac, ?begin, ?end - ?begin + 1) AS ?regionOfInterest)
            ?range
        WHERE {
            VALUES (?entry ?position) {""" + unpid_site_list_str + """}
            GRAPH <http://sparql.uniprot.org/uniprot> {
                ?entry up:sequence ?sequence .
                ?range faldo:begin
                   [ faldo:position ?begin ; faldo:reference ?sequence ] ;
                      faldo:end
                   [ faldo:position ?end ; faldo:reference ?sequence ] .
                FILTER (?begin <= ?position && ?position <= ?end)
                ?sequence rdf:value ?iupac
            }
        }
    }
    ?entry up:annotation ?annotation .
    ?annotation up:range ?range ;
                a ?type .
	OPTIONAL {
        ?annotation rdfs:comment ?comment .
    }
}"""

    # Grabs the annotation by parts to maximize the amount we receive from the uniprot server
    #region_annot = request_annot(new_entry_length_pos_query)
    logging.debug("Starting entry_len_pos query")
    region_annot = request_annot(alt_query_entry_len_pos)
    #region_annot_stripped = request_annot(query_stripped_entry_pos)
    #if len(region_annot.index) < len(region_annot_stripped.index):  # This should check to see if there are rows missing in the full query
    #    region_annot = region_annot.append(region_annot_stripped)
    #    region_annot.drop_duplicates(subset=["entry", " position", " begin", " end"], keep="first", inplace=True)
    logging.debug("Starting subcellular query")
    subcell_annot = request_annot(query_entry_subcellular)
    logging.debug("Starting ec_rhea_type query")
    extras_annot = request_annot(query_ec_rhea_type)
    logging.debug("Starting comment query")
    comment_annot = request_annot(alt_query_comment)  # entry, position, type, comment, begin, end, regionOfInterest
    # Creates blank dataframes if uniprot did not return that info
    if region_annot is None or list(region_annot.columns) != ["entry", " position", " lengthOfSequence"]:
        logging.debug(f"If the region header is similar to expected, Uniprot may have changed their database.\n{str(list(region_annot.columns))}")
        #columns=["entry", " position", " lengthOfSequence", " begin", " end", " regionOfInterest"]
        region_annot = pd.DataFrame(data = {"entry": pd.Series([], dtype="object"),
                                            " position": pd.Series([], dtype="object"),
                                            " lengthOfSequence": pd.Series([], dtype="object")})
    else:
        logging.debug(f"region data types are {region_annot.dtypes}")
        region_annot.dropna(how="all", inplace=True)

    if subcell_annot is None or list(subcell_annot.columns) != ["entry", " location"]:
        logging.debug(f"If the subcell header is similar to expected, Uniprot may have changed their database.\n{str(list(subcell_annot.columns))}")
        subcell_annot = pd.DataFrame(data = {"entry": pd.Series([], dtype="object"), 
                                             " location": pd.Series([], dtype="object")})
    else:
        logging.debug(f"subcell data types are {subcell_annot.dtypes}")
        subcell_annot.dropna(how="any", inplace=True)

    if extras_annot is None or list(extras_annot.columns) != ["entry", " position", " ec", " rhea"]:
        logging.debug(f"If the extras header is similar to expected, Uniprot may have changed their database.\n{str(list(extras_annot.columns))}")
        extras_annot = pd.DataFrame(data = {"entry": pd.Series([], dtype="object"), 
                                            " position": pd.Series([], dtype="object"), 
                                            " ec": pd.Series([], dtype="object"), 
                                            " rhea": pd.Series([], dtype="object")})
    else:
        logging.debug(f"extras data types are {extras_annot.dtypes}")
        extras_annot.dropna(how="all", inplace=True)  # TODO: I will likely need to change this back to all later
    
    if comment_annot is None or list(comment_annot.columns) != ["entry", " position", " type", " comment", " begin", " end", " regionOfInterest"]:
        logging.debug(f"If the comment header is similar to expected, Uniprot may have changed their database.\n{str(list(comment_annot.columns))}")
        comment_annot = pd.DataFrame(data={"entry": pd.Series([], dtype="object"), 
                                           " position": pd.Series([], dtype="object"), 
                                           " type": pd.Series([], dtype="object"), 
                                           " comment": pd.Series([], dtype="object"),
                                            " begin": pd.Series([], dtype="object"),
                                            " end": pd.Series([], dtype="object"),
                                            " regionOfInterest": pd.Series([], dtype="object")})
    else:
        logging.debug(f"comment data types are {comment_annot.dtypes}")
        comment_annot.dropna(how="all", inplace=True)

    # Full outer joins the annotations to preserve all info possible
    full_annot = region_annot
    del(region_annot)
    try:
        full_annot = full_annot.merge(subcell_annot, how="outer", on="entry")
        del(subcell_annot)
        full_annot = full_annot.merge(extras_annot, how="outer", on=["entry", " position"])
        del(extras_annot)
        full_annot = full_annot.merge(comment_annot, how="outer", on=["entry", " position"])
        del(comment_annot)
    except KeyError:  # The proper columns are missing in one of the query results; usually means uniprot isn't working correctly
        #print(full_annot)
        if isinstance(full_annot, pd.DataFrame) and "502 Proxy Error" in full_annot.iloc[1][0]:
            logging.exception("502 Error")
            print("502 ERROR: try Uniprot annotations later")
            return 502
        logging.exception("Uniprot data formatting error: See sparql_request function in sparql.py")
        print("Uniprot data formatting error: please notify ProteoSushi administrator")
        print(list(full_annot.columns))
        sys.exit()
    try:  # ['entry', ' position', ' lengthOfSequence', ' location', ' ec', ' rhea', ' type', ' comment', ' begin', ' end', ' regionOfInterest']
        # [0, 1, 2, 8, 9, 10, 3, 4, 5, 6, 7]
        full_annot = full_annot[["entry"," position"," lengthOfSequence"," begin",
                                                 " end"," regionOfInterest"," location"," ec",
                                                 " rhea"," type"," comment"]]
        #sparql_from_csv_df = sparql_from_csv_df[[0, 1, 2, 8, 9, 10, 3, 4, 5, 6, 7]]
    except Exception as e:
        logging.debug(f"Couldn't reorder columns due to error {e}")
    return full_annot

def request_annot(query: str, attempts_left=10):
    """Sends the query to uniprot and checks the output

    Arguments:
        query {str} -- the query with the sites to annotate
        attempts_left {int} -- the number of requests that can be sent out
    Returns:
        pandas.dataframe -- the returned annotation
    """
    headers = {"user-agent": "rseymour@wustl.edu"}
    time_to_sleep = 5
    try:
        logging.debug(f"{datetime.now()} -- Attempting post to sparql.uniprot.org")
        r = requests.post(ENDPOINT, data = {"format": "csv", "query": query}, headers = headers)
        if "<!DOCTYPE html SYSTEM \"about:legacy-compat\">" in r.text:  # This just means it returned a 500 error
            logging.warning("Query had a 500 error")
            if attempts_left > 0:
                print(r.text)
                sleep(time_to_sleep)
                print("\033[93m {}\033[00m".format(f"\nAttempts used: {11 - attempts_left}"), end='')
                return request_annot(query, attempts_left - 1)
            else:
                return None  # TODO: remove this once Uniprot fixes their stuff
                raise pd.errors.ParserError("Ran out of Uniprot accession attempts")
        logging.debug(f"{datetime.now()} -- Received response from sparql.uniprot.org")
        #logging.debug(f"Output is {r.text}")
        csv_file = StringIO(r.text)
        #logging.debug(f"CSV file is {csv_file}")
        # channels the data into a pandas dataframe and sorts it on the computer
        unsorted_pd_df = pd.read_csv(csv_file, encoding="utf-8", sep=',')
        try:
            sparql_from_csv_df = unsorted_pd_df.sort_values(by=["entry", " position"], ascending=[False, True])
        except KeyError:
            sparql_from_csv_df = unsorted_pd_df.sort_values(by=["entry"], ascending=False)
    except (pd.errors.ParserError, requests.exceptions.ConnectionError):
        logging.warning("Query had a connection/parser error")
        if attempts_left > 0:
            sleep(time_to_sleep)
            print("\033[93m {}\033[00m".format(f"\nAttempts used: {11 - attempts_left}"), end='')
            return request_annot(query, attempts_left - 1)
        else:
            return None  # TODO: remove this once Uniprot fixes their stuff
            raise pd.errors.ParserError("Ran out of Uniprot accession attempts")
    return sparql_from_csv_df

def process_sparql_output(output_df, sparql_dict: dict) -> list:
    """Process the output from uniprot to make it consistent with the rollup output

    Arguments:
        output_df {pandas.dataframe} -- the string output from uniprot, may have multiple
        sparql_dict {dict} -- the growing dictionary of uniprot annotations
    Returns:
        list -- a list of annotations with the data properly separated
    """
    #print(f"\nIn process_sparql_output beginning uniprot annotation is {type(output_df)}")
    logging.debug(f"\nIn process_sparql_output beginning uniprot annotation is {type(output_df)}")

    def __parse_sparql_line(line: str) -> str:
        """changes the complicated uniprot annotation output into a cleaner line

        Arguments:
            line {str} -- a line from the output
        Returns:
            str -- a cleaned up version of that same line
        """
        #entry,lengthOfSequence,catalyicActivity,location,ec,rhea,type,comment,position,begin,end,regionOfInterest
        new_str = list()
        for chunk in line[1]:
            if isinstance(chunk, str) and "http://purl.uniprot.org/uniprot/" in chunk:
                new_str.append(chunk.replace("http://purl.uniprot.org/uniprot/", ""))
            elif isinstance(chunk, str) and "http://purl.uniprot.org/core/" in chunk:
                new_str.append(chunk.replace("http://purl.uniprot.org/core/", ""))
            elif isinstance(chunk, str) and "http://purl.uniprot.org/locations/" in chunk:
                new_chunk = chunk.replace("http://purl.uniprot.org/locations/", "")
                try:
                    new_str.append(subcellular_location_dict[new_chunk.zfill(4)])  # Converts the number location to descriptor
                except KeyError:
                    new_str.append("")
            elif isinstance(chunk, str) and "http://purl.uniprot.org/enzyme/" in chunk:
                new_chunk = chunk.replace("http://purl.uniprot.org/enzyme/", "")
                new_str.append(ec_id_to_description_dict[new_chunk])
            elif isinstance(chunk, str) and "^^<http://www.w3.org/2001/XMLSchema#int>" in chunk:
                new_str.append(findall(r"(\d+?)\^", chunk)[0])
            else:
                new_str.append(str(chunk))
        return '\t'.join(new_str)

    output_list = list()
    comments_dict = dict()
    #sparql_dict = dict()  # This gets used by the main program to connect these annotations to the rest of the data.
    if len(output_df) == 1 or (not isinstance(output_df, str) and output_df.empty):
        print("\nUniprot Error: Failed to get annotations", end='')
        return output_list, sparql_dict
    
    try:
        position_index = output_df.columns.get_loc(" position")
    except (AttributeError, KeyError):
        logging.exception("Uniprot annotations missing position column or wrong format")
        print("\nUniprot Error: Missing the Position column", end='')
        logging.debug(f"In process_sparql_output error uniprot annotation is {type(output_df)}")
        return output_list, sparql_dict

    # get the subcellular location from UniProt
    def retrieve_uniprot_subcellular_location(attempts_left=5):
        '''Retrieve subcellular location information from UniProt directly'''
        if attempts_left <= 0:
            return 4
        try:
            r = requests.get('https://www.uniprot.org/locations/?format=tab')
        except requests.exceptions.ConnectionError:
            return retrieve_uniprot_subcellular_location(attempts_left-1)
        subcellular_location_df = pd.read_csv(StringIO(r.text), sep = '\t')  # TODO: potential ParserError
        try:
            subcellular_location_df["location_id_reformatted"] = (
                                                                subcellular_location_df["Subcellular location ID"]
                                                                .str.split("-")
                                                                .str[1]
                                                                )
        except KeyError:
            subcellular_location_df["location_id_reformatted"] = ("")
        return subcellular_location_df
    
    subcellular_location_df = retrieve_uniprot_subcellular_location()
    if isinstance(subcellular_location_df, int) and subcellular_location_df == 4:
        return 4, None  # This is an error related to uniprots's 
    subcellular_location_dict = dict(subcellular_location_df[["location_id_reformatted", "Alias"]].values.tolist())

    # Fetches a file from expasy to deal with the 'ec' column
    try:
        r = urllib.request.urlopen("ftp://ftp.expasy.org/databases/enzyme/enzyme.dat", "enzyme.dat")
    except urllib.error.URLError:
        sleep(10)
        r = urllib.request.urlopen("ftp://ftp.expasy.org/databases/enzyme/enzyme.dat", "enzyme.dat")
    enzyme_dat_list = r.read().decode("utf-8").split('\n')

    # Turn enzyme.dat into a dictionary for mapping
    ec_id_to_description_dict = dict()
    for line_number, line in enumerate(enzyme_dat_list):
        if line.startswith("ID "):
            current_id = line[5:]
            current_description = enzyme_dat_list[line_number + 1][5:]
            ec_id_to_description_dict[current_id] = current_description

    #entry,position,lengthOfSequence,location,ec,rhea,type,comment,begin,end,regionOfInterest
    #header = output_lines.colnames()
    #print(output_lines.columns)
    
    #i = 1
    repeat_index = 3
    #while i < len(output_lines):
    for output_line in output_df.iterrows():  # TODO: optimize by vectorizing
        output_line = __parse_sparql_line(output_line)
        # This should only have 1 iteration and is used to properly split the tsv type line
        for output_split in csv.reader([output_line], delimiter='\t', quotechar='"'):
            key = ','.join([*output_split[:repeat_index]])
            try:
                comments_dict[key] += output_split[repeat_index:]
            except KeyError:
                comments_dict[key] = output_split[repeat_index:]
        #i += 1
    
    for key in comments_dict:
        output_list.append(key + ',' + ','.join(comments_dict[key]))
        sparql_dict[ key.split(',')[0] + '|' + key.split(',')[1] ] = key.split(',') + comments_dict[key] #unpid, pos
    #print(f"sparql_dict has {sparql_dict}")
    return output_list, sparql_dict
#EOF


"""ps_utilities.py: helper functions used by parsers in proteoSushi"""

__author__ = "Arshag Mooradian, Rob Seymour"
__email__ = "rseymour@wustl.edu"

import pickle
import csv
from collections import defaultdict
import ntpath
import os
from re import finditer
try:
    from .ps_digest import digest, fasta_producer, cleave_rule_determination
except ImportError:
    from ps_digest import digest, fasta_producer, cleave_rule_determination


def clean_localization_pep_seq(rule: tuple, pep_mod_seqs: list, user_PTMs: list, old_pep_seq: str) -> tuple:
    """Removes the missed cleavages without selected PTMs

    Arguments:
        rule {tuple} -- contains info about the protease
        pep_mod_seq {str} -- the peptide sequence with modifications
        user_PTMs {list} -- a list of PTMs that the user has selected for Analysis
    Returns:
        list -- the cleaned PTM localization peptide
        str -- the cleaned peptide
        int -- length of the front part of peptide that was removed
    """
    # Puts break points for unmodified sequence in a list
    breaks = finditer(rule[0], old_pep_seq)
    cut_unmod_sites = []
    cut_unmod_peptides = []
    for breakp in breaks:
        cut_unmod_sites.append(breakp.start())
    # Adds the unmodified cut peptides in one at a time
    last_site = 0
    if rule[1].lower() == 'c':
        for i in cut_unmod_sites:
            cut_unmod_peptides.append(old_pep_seq[last_site:i+1])
            last_site = i + 1
        cut_unmod_peptides.append(old_pep_seq[last_site:])
    elif rule[1].lower() == 'n':
        for i in cut_unmod_sites:
            cut_unmod_peptides.append(old_pep_seq[last_site:i])
            last_site = i
        cut_unmod_peptides.append(old_pep_seq[last_site:])
    #if old_pep_seq == "KACLNPASPIVK":
    #    input("peptide split " + str(cut_unmod_peptides))

    PTM_cut_peptides = []
    for pep_mod_seq in pep_mod_seqs:
        if pep_mod_seq == '':
            PTM_cut_peptides.append('')
            continue
        # Puts the break points for modified sequence in a list
        breaks = finditer(rule[0], pep_mod_seq)
        cut_sites = []
        cut_peptides = []
        for breakp in breaks:
            cut_sites.append(breakp.start())
        # Adds the cut peptides in one at a time
        last_site = 0
        if rule[1].lower() == 'c':
            for i in cut_sites:
                cut_peptides.append(pep_mod_seq[last_site:i+1])
                last_site = i + 1
            cut_peptides.append(pep_mod_seq[last_site:])
        elif rule[1].lower() == 'n':
            for i in cut_sites:
                cut_peptides.append(pep_mod_seq[last_site:i])
                last_site = i
            cut_peptides.append(pep_mod_seq[last_site:])
        PTM_cut_peptides.append(cut_peptides)
    #if old_pep_seq == "KACLNPASPIVK":
    #    input("peptide split " + str(PTM_cut_peptides))
    try:
        cut_peptides = [s for s in PTM_cut_peptides if s != ""][0]
    except IndexError:
        return None, None, None
    #if old_pep_seq == "KACLNPASPIVK":
    #    print(cut_peptides)
    first_start = len(cut_peptides)
    last_end = 0
    for cut_peptides in [s for s in PTM_cut_peptides if s != ""]:
    # This gets the range of the first and last pep_slice with a PTM
        start = 0
        while start < len(cut_peptides):
            if '(' in cut_peptides[start] or '[' in cut_peptides[start]:
                break
            else:
                start += 1
        if start < first_start:
            first_start = start

        end = len(cut_peptides) - 1
        while end > start:
            if '(' in cut_peptides[end] or '[' in cut_peptides[end]:
                break
            else:
                end -= 1
        if end > last_end:
            last_end = end
    #if old_pep_seq == "KACLNPASPIVK":
    #    input("start end " + str(first_start) + ' ' + str(last_end))
    if first_start == last_end:
        new_pep_mod_seqs = [peptide[first_start] for peptide in PTM_cut_peptides if peptide != '']
        new_pep_seq = cut_unmod_peptides[first_start]
    else:
        new_pep_mod_seqs = [''.join(peptide[first_start:last_end+1]) for peptide in PTM_cut_peptides if peptide != '']
        new_pep_seq = ''.join(cut_unmod_peptides[first_start:last_end+1])
    return new_pep_mod_seqs, new_pep_seq, len(''.join(cut_unmod_peptides[0:first_start]))

def clean_pep_seq(rule: tuple, pep_mod_seq: str, user_PTMs: list, old_pep_seq: str, is_MQ = False) -> str:
    """Removes the missed cleavages without selected PTMs

    Arguments:
        rule {tuple} -- contains info about the protease
        pep_mod_seq {str} -- the peptide sequence with modifications
        user_PTMs {list} -- a list of PTMs that the user has selected for Analysis
        old_pep_seq {str} -- the unmodified peptide sequence
        is_MQ {bool} -- whether the search engine is MQ
    Returns:
        str -- the cleaned peptide
    """
    # Puts break points for unmodified sequence in a list
    breaks = finditer(rule[0], old_pep_seq)
    cut_unmod_sites = []
    cut_unmod_peptides = []
    for breakp in breaks:
        cut_unmod_sites.append(breakp.start())
    # Adds the unmodified cut peptides in the list one at a time
    last_site = 0
    if rule[1].lower() == 'c':
        for i in cut_unmod_sites:
            cut_unmod_peptides.append(old_pep_seq[last_site:i+1])
            last_site = i + 1
        cut_unmod_peptides.append(old_pep_seq[last_site:])
    elif rule[1].lower() == 'n':
        for i in cut_unmod_sites:
            cut_unmod_peptides.append(old_pep_seq[last_site:i])
            last_site = i
        cut_unmod_peptides.append(old_pep_seq[last_site:])

    # Puts the break points for modified sequence in a list
    breaks = finditer(rule[0], pep_mod_seq)
    cut_sites = []
    cut_peptides = []
    for breakp in breaks:
        cut_sites.append(breakp.start())
    # Adds the cut peptides in the list one at a time
    last_site = 0
    if rule[1].lower() == 'c':
        for i in cut_sites:
            cut_peptides.append(pep_mod_seq[last_site:i+1])
            last_site = i + 1
        cut_peptides.append(pep_mod_seq[last_site:])
    elif rule[1].lower() == 'n':
        for i in cut_sites:
            cut_peptides.append(pep_mod_seq[last_site:i])
            last_site = i
        cut_peptides.append(pep_mod_seq[last_site:])

    # This gets the range of the first and last pep_slice with a PTM
    start = 0
    while start < len(cut_peptides):
        if is_MQ:
            if any(user_PTM[:2].lower() in cut_peptides[start] for user_PTM in user_PTMs):
                break
            else:
                start += 1
        else:
            if any(user_PTM in cut_peptides[start] for user_PTM in user_PTMs):
                break
            else:
                start += 1
    
    end = len(cut_peptides) - 1
    while end > start:
        if is_MQ:
            if any(user_PTM[:2].lower() in cut_peptides[end] for user_PTM in user_PTMs):
                break
            else:
                end -= 1
        else:
            if any(user_PTM in cut_peptides[end] for user_PTM in user_PTMs):
                break
            else:
                end -= 1

    if start == end:
        new_pep_mod_seq = cut_peptides[start]
        new_pep_seq = cut_unmod_peptides[start]
    else:
        new_pep_mod_seq = ''.join(cut_peptides[start:end+1])
        new_pep_seq = ''.join(cut_unmod_peptides[start:end+1])
    new_start = start
    while len(new_pep_seq) < 6:
        if new_start <= 0:
            end += 1
            new_pep_mod_seq = ''.join(cut_peptides[new_start:end+1])
            new_pep_seq = ''.join(cut_unmod_peptides[new_start:end+1])
        elif end >= len(cut_unmod_peptides) - 1:
            new_start -= 1
            new_pep_mod_seq = ''.join(cut_peptides[new_start:end+1])
            new_pep_seq = ''.join(cut_unmod_peptides[new_start:end+1])
        elif len(cut_unmod_peptides[new_start-1]) >= len(cut_unmod_peptides[end+1]):
            new_start -= 1
            new_pep_mod_seq = ''.join(cut_peptides[new_start:end+1])
            new_pep_seq = ''.join(cut_unmod_peptides[new_start:end+1])
        else:
            end += 1
            new_pep_mod_seq = ''.join(cut_peptides[new_start:end+1])
            new_pep_seq = ''.join(cut_unmod_peptides[new_start:end+1])
    #input(f"old {pep_mod_seq}, new {new_pep_mod_seq}")
    return new_pep_mod_seq, new_pep_seq, len(''.join(cut_unmod_peptides[0:start]))

def parse_maxquant_summary(infile: str) -> list:
    """Pulls out information needed from the MaxQuant files
    Arguments:
        infile {str} -- the name of the MaxQuant summary file that will be used
    Returns:
        str -- the name of the protease used (usually trypsin)
        int -- the maximum allowed number of missed cleavages
    """
    mqfile = open(infile, 'r')
    tsvReader = csv.reader(mqfile, delimiter='\t', quotechar='"')
    protease = ""
    PTMs = []
    maxMissedCleaves = 0
    # From proteinGroups: Need Intensity[DM], PTM data[GS-GX], Potential Contaminant(?)[GK]
    # From summary: protease[C], Variable Modifications[H][I-J?], Max. missed cleavages[O]
    # From parameters: Modifications included in protein quantification,
    header = next(tsvReader)  # Skips the header line
    row = next(tsvReader)
    protease = row[header.index("Enzyme")]
    assert protease != ""
    maxMissedCleaves = int(row[header.index("Max. missed cleavages")])
    mqfile.close()
    return protease.lower(), maxMissedCleaves

def parse_mascot(in_file: str) -> list:
    """Pulls out protease, quantitative range, and variable modifications, along
    with the quant data from a Mascot file.
    Arguments:
        in_file {str} -- a string for the name of the mascot file
    Returns:
        list -- protease {str}, quant_range {tuple}, var_mod_map {dict},
        missed_cleaves {int}
    """
    # We need to parse Mascot file first, because raw Mascot output is difficult to work with
    print(f"Parsing file {in_file}.")
    with open(in_file) as f, \
        open('quant_parsed.csv', 'w', newline='') as out_file:
        # Quantitative section of Mascot file
        assert in_file.endswith('.csv'), "File must be a csv"
        file_reader = csv.reader(f)
        out_writer = csv.writer(out_file)
        protease = ""
        missed_cleaves = -1
        quant_range = None
        var_mod_map = dict()
        for row in file_reader:
            if row:  # lots of blank rows
                # Pull data we need
                if row[0] == "Enzyme":
                    protease = row[1].lower()
                elif row[0] == "Variable modifications" and '--' not in row[1]:
                    var_mods = row[1].split(',')
                    var_mod_map = {val.lower():
                                   str(pos+1) for pos, val in enumerate(var_mods)}
                elif row[0] == "Maximum Missed Cleavages":
                    missed_cleaves = int(row[1])
                elif row[0] == "prot_hit_num" or row[0] == "prot_acc":
                    # This should be last elif; start of quantitative info
                    header = row
                    header = [i for i in header if i]
                    second_row = next(file_reader)  # need to normalize headers
                    quant_range = (len(header), len(second_row)-1)  #range where there will be quant; appears to skip some
                    header.extend(['SampleQuant' 
                                   for i in range(len(second_row) - len(header))])
                    out_writer.writerows([header, second_row])
                    for row in file_reader:
                        out_writer.writerow(row)
                    break  # Exit loop as we have all of the header information
        print("Finished Parsing")
        if quant_range is None:
            return -5, None, None, None
        return protease, quant_range, var_mod_map, missed_cleaves


def load_pepdict(proteome_fasta_filepath: str, protease: str, missed_cleavages: int) -> dict:
    """Looks for pep_dict; if it doesn't find, it looks for a FASTA to make one.
    Fasta would be in current directory but this can be changed.

    Arguments:
        proteome_fasta_filepath {str} -- the filepath for the proteome fasta file
        protease {str} -- protease used in the sample digestion
        missed_cleavages {int} -- the max number of allowed missed cleavages
    Returns:
        dict -- a dictionary that connects peptide sequence to gene, position, protein name and unpid
    """
    pep_dict = None
    #wd = os.getcwd()
    fasta_dir = os.path.dirname(os.path.abspath(proteome_fasta_filepath))
    pep_dict_file = f"{proteome_fasta_filepath.split('.')[0]}_{protease.replace('/','^')}.pepdict"
    #print(os.listdir(fasta_dir))
    if not any([x == ntpath.basename(pep_dict_file) for x in os.listdir(fasta_dir)]):
        print("Pepdict not found in the FASTA directory. Trying to generate from the FASTA file")
        digest_if_needed(proteome_fasta_filepath, protease, missed_cleavages)
    
    print(f"Loading {pep_dict_file} into memory")
    with open(pep_dict_file, 'rb') as f:
        pep_dict = pickle.load(f)
    print("\033[95m {}\033[00m".format("Peptide Dictionary Loaded."))
    return pep_dict


def digest_if_needed(proteome_fasta_filepath: str, protease: str, missed_cleavages: int, minlen=6, maxlen=55, 
                     m_excise="Both"):
    """Runs digest and dumps pickle file (dict)"""
    seq_list = []  # (gene, organism, sequence, unpid)
    seq_list = fasta_producer(proteome_fasta_filepath, seq_list)
    rule = cleave_rule_determination(protease)
    pep_dict = defaultdict(set)  # NOTE: this won't work for large fasta/sequence databases as it's in ram; ~1GB for human
    for i in seq_list:
        gene, _organism, seq, unpid, protein_name = i
        if m_excise == "Both":
            # Run digest results twice and combine; inefficient but gets job done
            digest_results_true = digest(seq, minlen, maxlen, missed_cleavages, True,
                                         li_swap=True, rule_to_use=rule,
                                         reverse=False)
            digest_results_false = digest(seq, minlen, maxlen, missed_cleavages, False,
                                          li_swap=True, rule_to_use=rule,
                                          reverse=False)
            digest_results = set().union(digest_results_true,
                                         digest_results_false)
        else:
            digest_results = digest(seq, minlen, maxlen, missed_cleavages, m_excise,
                                    li_swap=True, rule_to_use=rule,
                                    reverse=False)
        for result in digest_results:
            peptide, position = result
            peptide = peptide.replace("L", "I")
            pep_dict[peptide].add((gene, position, unpid, protein_name))  # careful with site specificity here
    with open(f"{proteome_fasta_filepath.split('.')[0]}_{protease.replace('/','^')}.pepdict", "wb") as w1:
        pickle.dump(pep_dict, w1)
    del pep_dict  # Free memory for later



import codecs
import setuptools

#with open("README.md", "r") as fh:
with codecs.open("README.md", mode='r', encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="proteosushi",
    version="1.5.1",
    author="Rob Seymour",
    author_email="rseymour@wustl.edu",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HeldLab/ProteoSushi",
    download_url="https://github.com/HeldLab/ProteoSushi/archive/v1.1.0.tar.gz",
    include_package_data=True,
    packages=setuptools.find_packages(),
    package_dir={
        "lib": "proteosushi",
        #"": "lib"
    },
    package_data={
        "lib": ['spec_list_fixed.tsv']
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: X11 Applications :: Qt",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    python_requires=">=3.7",
    install_requires=[
        "PyQt5>=5.15",
        "requests>=2.24",
        "pandas>=1.0"
    ]
)


"""run_sparql.py: runs the sparql script independent of ProteoSushi"""

try:
    from .sparql import sparql_request
except ImportError:
    from sparql import sparql_request

from time import sleep

'''with open("sparqllist.tsv", 'r') as sparqllist:
    sparql_input = sparqllist.readlines()
    batch = 1
    i = 0
    while i + batch <= len(sparql_input):
        print(sparql_input[i:i+batch])
        sparql_output = sparql_request(sparql_input[i:i+batch])
        i += batch
        print(f"{float(i)/len(sparql_input)*100}% of results annotated")
        sleep(1)
        print(sparql_output.text)'''

sparql_input = ["(uniprotkb:O00468 317)",
    "(uniprotkb:O00468 948)",
    "(uniprotkb:O00468 844)",
    "(uniprotkb:O00468 103)",
    "(uniprotkb:O00468 296)",
    "(uniprotkb:O00468 1113)",
    "(uniprotkb:O00468 750)",
    "(uniprotkb:O00468 285)",
    "(uniprotkb:O00468 440)",
    "(uniprotkb:O00468 847)",
    "(uniprotkb:O00468 399)",
    "(uniprotkb:P23526 195)",
    "(uniprotkb:O43865 211)",
    "(uniprotkb:O43865 172)",
    "(uniprotkb:O43865 166)",
    "(uniprotkb:O43865 151)",
    "(uniprotkb:O43865 373)",
    "(uniprotkb:Q09666 1833)",
    "(uniprotkb:P02765 219)",
    "(uniprotkb:P02765 132)",
    "(uniprotkb:P02765 230)",
    "(uniprotkb:Q9Y4K1 503)",
    "(uniprotkb:Q12904 161)",
    "(uniprotkb:Q13155 306)",
    "(uniprotkb:O00170 208)",
    "(uniprotkb:O00170 90)",
    "(uniprotkb:P54819 92)",
    "(uniprotkb:P14550 134)",
    "(uniprotkb:P14550 46)",
    "(uniprotkb:P15121 187)",
    "(uniprotkb:P15121 200)",
    "(uniprotkb:P15121 299)",
    "(uniprotkb:P15121 304)",
    "(uniprotkb:Q04828 188)",
    "(uniprotkb:Q04828 206)",
    "(uniprotkb:Q04828 242)",
    "(uniprotkb:Q04828 193)",
    "(uniprotkb:P42330 242)",
    "(uniprotkb:O43488 132)",
    "(uniprotkb:O43488 156)",
    "(uniprotkb:O95154 186)",
    "(uniprotkb:P31749 296)",
    "(uniprotkb:P02768 500)",
    "(uniprotkb:P02768 289)",
    "(uniprotkb:P02768 501)",
    "(uniprotkb:Q13740 313)",
    "(uniprotkb:Q13740 392)",
    "(uniprotkb:Q13740 354)",
    "(uniprotkb:Q13740 270)",
    "(uniprotkb:Q8IZ83 350)"]

sparql_output = sparql_request(sparql_input)
print(sparql_output)
#EOF